﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CommunityDetection
{
	public class DisjunctionBuilder : Builder
	{
		HashSet<int> allRows;
		double alpha;
		int maxNumPatterns;
		int beamWidth;
		int maxPatternLength;

		public DisjunctionBuilder(BinaryMatrix data, int maxNumPatterns, int beamWidth, int maxPatternLength, double alpha) : base(data)
		{
			allRows = new HashSet<int>();
			for (int i = 0; i < data.M; i++)
				allRows.Add(i);
			this.maxNumPatterns = maxNumPatterns;
			this.beamWidth = beamWidth;
			this.maxPatternLength = maxPatternLength;
			this.alpha = alpha;
		}

		public override Classifier build(HashSet<int> trueRows)
		{
			HashSet<int> rows = new HashSet<int>(allRows);
			Dictionary<int, int> attributes = new Dictionary<int,int>();
			foreach(int r in trueRows)
			{
				foreach (int a in data.getRow(r).getIndexsByValue(true))
				{
					if (!attributes.ContainsKey(a))
						attributes[a] = 0;
					attributes[a]++;
				}
			}

			List<DisjunctionPattern> list = new List<DisjunctionPattern>();
			for (int i = 0; i < maxNumPatterns; i++)
			{
				//Console.WriteLine(attributes.Count + " attributes");
				DisjunctionPattern pattern = findBestPattern(rows, attributes, trueRows);
				if (pattern.pattern.Count == 0)
				{
					//Console.WriteLine("Not possible");
					break;
				}

				int d = rows.Count;
				foreach (int r in pattern.rows)
				{
					if (!trueRows.Contains(r))
						continue;
					BitArray row = data.getRow(r);
					//foreach (int j in row.getIndexsByValue(true))
					{
						//attributes[j]--;
						//if (attributes[j] == 0)
						//	attributes.Remove(j);
					}
				}
				//Console.WriteLine(attributes.Count + " attributes");
				trueRows.RemoveWhere(x => pattern.rows.Contains(x));
				rows.RemoveWhere(x => pattern.rows.Contains(x));
				d -= rows.Count;
				//Console.Write("Pattern: " + pattern+", rlt: "+rows.Count+", trl: "+trueRows.Count+", d:"+d);
				list.Add(pattern);
			}
			return new DisjunctionClassifier(list);
		}

		public double measure(double tp, double fp, double tn, double fn)
		{
			double p = tp / (tp + fp);
			double r = tp / (tp + fn);
			double v=  1.0 / (alpha * (1.0 / r) + (1.0 - alpha)*(1.0/p));
			//Console.WriteLine(v);
			return v;
		}
		public DisjunctionPattern findBestPattern(HashSet<int> rows, Dictionary<int,int> attributes, HashSet<int> trueRows)
		{
			//init
			int depth = 0;
			List<DisjunctionPattern> beam = new List<DisjunctionPattern>();
			DisjunctionPattern best = new DisjunctionPattern();
			best.rows = new HashSet<int>(rows);
			beam.Add(best);

			while (depth < maxPatternLength)
			{
				HashSet<DisjunctionPattern> newBeam = new HashSet<DisjunctionPattern>();
				foreach (DisjunctionPattern pattern in beam)
				{
					foreach (int a in attributes.Keys)
					{
						//check if att is part of pattern
						if (pattern.pattern.Contains(a))
							continue;

						//create new pattern
						DisjunctionPattern newpat = new DisjunctionPattern();
						newpat.AddAttribute(a);
						foreach (int pa in pattern.pattern)
							newpat.AddAttribute(pa);
						if (newBeam.Contains(newpat))
							continue;

						double tp = 0; //contains pattern, class 1
						double fp = 0; //contains pattern, class 0
						double tn = 0; //does not contain pattern;
						double fn = 0; //

						foreach (int r in pattern.rows) //only check pattern rows, other rows cant contain new pattern because it is superset of "pattern"
						{
							if (data.getRow(r).contains(newpat.pattern))
							{
								newpat.AddRow(r);
								if (trueRows.Contains(r))
									tp++;
								else
									fp++;
							}
							else
							{
								if (trueRows.Contains(r))
									fn++;
								else
									tn++;
							}
						}
						//Console.WriteLine(tp + " " + fp + " " + fn + " " + tn);

						//calculate information gain
						newpat.gain = measure(tp, fp, tn, fn);

						//add to new beam
						newBeam.Add(newpat);
					}
				}

				//sort new beam by information gain
				beam = new List<DisjunctionPattern>(newBeam);
				beam.Sort();

				//best one improved?
				if (best.gain < beam[0].gain)
				{
					best = beam[0];
				}
				beam.RemoveRange(Math.Min(beam.Count, beamWidth) - 1, Math.Max(beam.Count - beamWidth, 0));
				depth++;
			}
			return best;
		}
	}
}
