﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CommunityDetection
{
	/**
	 * Just a datasctructure to hold the best pattern seen till now
	 */
	public class Best
	{
		public double ig;
		public List<int> pattern;

		public Best()
		{
			ig = 0;
			pattern = null;
		}

		public override string ToString()
		{
			string best = pattern == null ? "gain:" + ig + ", pattern: null" : "gain:" + ig + ", pattern:";
			if (pattern != null)
				foreach (int i in pattern)
					best += " " + i;
			return best;
		}
	}

	/**
	 * Just a datastructure to remember a information gain score and the corresponding conditional database
	 */
	public class IG
	{
		public List<int> conditionalDB;
		public double ig;

		public override string ToString()
		{
			return "IG: "+ig;
		}
	}

	class Partition
	{

		//best pattern seen till now
		readonly Best bestPattern;

		//indices of rows belonging to this pattern
		public List<int> partitionRows;

		//description data matrix
		BinaryMatrix data;

		//set of all rows (whole dataset, not only in partition) with classlabel 1
		HashSet<int> trueRows;

		//used for splitting
		public PatternNode splitNode;

		//minimum support
		int minsup = 1;

		//P( c=1 )
		public double p;

		//entropy of partitio
		public double entropy;

		public Partition(BinaryMatrix data, HashSet<int> trueRows, List<int> rows)
		{
			//save references
			this.data = data;
			this.trueRows = trueRows;
			this.partitionRows = rows;

			//minimum support
			minsup = 1;

			//node used for the classifier, to save at the end on which pattern this partition is split
			splitNode = new PatternNode();
			
			//best pattern seen in branch-and-bound algorithm, only properties of this object are changed
			bestPattern = new Best();

			//calculate the fraction with classlabel 1 in this paritition
			p = 0;
			foreach (int r in partitionRows)
				if (trueRows.Contains(r))
					p++;
			p /= partitionRows.Count;

			//calculate entropy of this partition
			entropy = p == 0 || p == 1 ? 0 : (-p * log(p) - (1 - p) * log(1 - p));
		}

		/**
		 * Find the best Pattern in a Partition p
		 */
		public Best findBestPattern()
		{
			HashSet<int> use = new HashSet<int>();
			for (int i = 0; i < 4; i++)
			{
				use.Add(i);
			}
			//run bnb on all rows
			return branchAndBound(new List<int>(), partitionRows, use);

		}


		/**
		 * Build (conditional) FPTree on certain rows, but only use items that are not in condition
		 */
		public FPTree getFPTreeFromRows(List<int> rows, HashSet<int> condition)
		{
			//calculate sum
			List<KeyValuePair<int, int>> list = new List<KeyValuePair<int, int>>();
			int[] sum = new int[data.N];

			//calculate itemfrequencies in rows
			foreach (int r in rows)
			{
				List<int> indices = data.getIndexesByValuePerRow(r, true);
				foreach (int i in indices)
				{
					sum[i]++;
				}
			}
			
			//select right items and create KVpair and set of items that are frequent
			HashSet<int> frequent = new HashSet<int>();
			for (int i = 0; i < sum.Length; i++)
			{
				if (!condition.Contains(i)) //do not use items that are used in
				{
					if (sum[i] >= minsup)
					{
						list.Add(new KeyValuePair<int, int>(i, sum[i]));
						frequent.Add(i);
					}
				}
			}

			//Console.WriteLine("-=-=-=-=-=");
			//Console.WriteLine("prefix: " + Tools.listToString(condition));
			//sort according to item freqency and create map for positions
			list.Sort(delegate(KeyValuePair<int, int> x, KeyValuePair<int, int> y) { return y.Value.CompareTo(x.Value); });
			Dictionary<int, int> position = new Dictionary<int, int>();
			for (int i = 0; i < list.Count; i++)
			{
				//item i is stored and position[x] in the KVList
				position[list[i].Key] = i;
			}
		
			//create FPTree
			FPTree tree = new FPTree();
			
			//loop through all transactions
			foreach (int i in rows)
			{
				//get transaction
				List<int> t = data.getIndexesByValuePerRow(i, true);
				//Console.WriteLine("orig: "+Tools.listToString(t));
				//remove all items that are in the condition field or are not frequent
				t.RemoveAll(transactionItem => condition.Contains(transactionItem) || !frequent.Contains(transactionItem));

				//sort according to global item frequency
				t.Sort(delegate(int x, int y) { return list[position[y]].Value.CompareTo(list[position[x]].Value); });

				//Console.WriteLine("in: " + Tools.listToString(t));

				//insert in FPTree tree
				tree.insertTree(t);

			}
			return tree;
		}

		/**
		 * Branch and bound algorithm as described in "Direct discriminative pattern mining for effective classification"
		 * @param tree An FPTree
		 * @param minsup Minimum support
		 * @param prefix  The current featureset that will be explored
		 * @param best The best found pattern till now, just an object to remember the best pattern in recursion
		 * @param rows List of row indices in the dataset
		 * @param trueRows Set of row indices with classlabel one
		 * @return The most discriminative pattern in rows
		 */
		public Best branchAndBound(List<int> prefix, List<int> rows, HashSet<int> use)
		{
			//create conditional FPTree
			FPTree tree = getFPTreeFromRows(rows,new HashSet<int>(prefix));
			//tree.print();
			//no frequent itemset in rows/items
			if (tree.count == 1)
				return bestPattern;

			//loop throug header table of FPTree, start at bottom
			for(int i=tree.header.Count-1; i>=0; i--)
			{
				//get FPNode
				FPNode itemnode = tree.header[i];

				if (!use.Contains(itemnode.item))
					continue;

				//Create pattern (prefix v node)
				List<int> pattern = new List<int>(prefix);
				pattern.Add(itemnode.item);

				//calculate information gain (IG object is for storing conditional database of pattern as well, saves calculations
				IG igPattern = gain(pattern, rows);

				//if best one, save in bestPattern object
				if (igPattern.ig > bestPattern.ig)
				{
					bestPattern.pattern = pattern;
					bestPattern.ig = igPattern.ig;
				}

				//conditional database (created in gain function)
				List<int> rowsConditional = igPattern.conditionalDB;

				//calculate upperbound with size of conditional database with support of pattern - 1
				double upper = gainUB( (double) (rowsConditional.Count-1) / rows.Count);

				//w.WriteLine("pattern: " + Tools.listToString(pattern) + " | prefix: " + Tools.listToString(prefix));

				//possible to improve, so no pruning
				if (bestPattern.ig <= upper) 
				{
					HashSet<int> itemsToUse = new HashSet<int>();
					while (itemnode != null)
					{
						FPNode path = itemnode;
						while (path.item != -1)
						{
							itemsToUse.Add(path.item);
							path = path.parent;
						}
						itemnode = itemnode.link;
					}
					//go into recursion with pattern-conditional database
					branchAndBound(pattern, igPattern.conditionalDB, itemsToUse);
				}
			}
			return bestPattern;
		}

		/**
		 * Return the 2log of number
		 */
		public double log(double d)
		{
			return Math.Log(d, 2);
		}

		/** 
		 * Calculate the conditional entropy based on a given theta and q
		 * @param theta P( x=1) (support dataset)
		 * @param q P( c=1 | x=1 ) (supp
		 */
		public double conditionalEntropy(double theta, double q)
		{
			//formula for conditional entropy as given in "Discriminative Frequent Pattern Analasys for Effective Classification"
			//treat x log x with x = 0 as 0
			
			double hcx1,hcx2,hcx3;

			if (theta == 0 || q == 0 || q == 1)
				hcx1 = 0;
			else
				hcx1 = -theta * q * log(q) - theta * (1 - q) * log(1 - q);
			
			if (theta * q == p || theta == 1)
				hcx2 = 0;
			else
				hcx2 = (theta * q - p) * log((p - theta * q) / (1 - theta));

			hcx3 = (theta * (1 - q) - (1 - p)) * log(((1 - p) - theta * (1 - q)) / (1 - theta));
			
			//divide by 0, or log 0, always means there is some 0 * log(0) or * 0 * (x/0)
			if (double.IsNaN(hcx3))
				hcx3 = 0;

			return hcx1 + hcx2 + hcx3;
		}

		/**
		 * Calculate the information gain ig of a certain pattern for certain rows
		 * @param pattern List of items in the pattern
		 * @param rows List of row indices in the current dataset
		 * @param set of row indices that have class label one
		 * @return The information gain of this dataset
		 */
		public IG gain(List<int> pattern, List<int> rows)
		{
			//Console.WriteLine("try pattern: " + Tools.listToString(pattern));
			double cp = 0; //contains pattern, class 1
			double cn = 0; //contains pattern, class 0
			double np = 0; //not contains pattern, class 1
			double nn = 0; //not contains pattern, class 0
			
			//conditional database for pattern
			List<int> dbCon = new List<int>();

			//loop through rows
			foreach (int r in rows)
			{
				BitArray row = data.getRow(r);

				//contains pattern?
				bool contains = true;
				foreach (int i in pattern)
				{
					if (!row.get(i))
					{
						contains = false;
						break;
					}
				}
				
				//increment right counter
				if (contains)
				{
					if (trueRows.Contains(r))
						cp++;
					else
						cn++;

					//contains pattern, so add to conditional database
					dbCon.Add(r);
				}
				else
				{
					if (trueRows.Contains(r))
						np++;
					else
						nn++;
				}
			}

			//formula page  4 "Discriminative frequent pattern analysis for effective classification"
			double theta = (cn + cp) / partitionRows.Count; // support pattern
			double q = cp / (cn + cp); // P( c=1 | pattern)

			//gain datastructure
			IG gain = new IG();
			gain.conditionalDB = dbCon;

			//calculate gain
			gain.ig = entropy - conditionalEntropy(theta, q);
			return gain;
		}

		/**
		 * Calculate the upper bound for a given theta
		 */
		public double gainUB(double theta)
		{
			if (theta < Math.Min(1 - p, p)) //left side of graph
			{
				return entropy - Math.Min(conditionalEntropy(theta, 0), conditionalEntropy(theta, 1));
			}
			/*else if(theta > Math.Max(1 - p,p)) //right side of graph
			{
				return entropy - Math.Min(conditionalEntropy(theta, p / theta), conditionalEntropy(theta, 1 - (1 - p) / theta));
			}*/
			// then can improve till entropy
			return entropy;

		}

		/**
		 * Split this parition on a certain pattern
		 */
		public List<Partition> split(Pattern pattern)
		{
			//split according to new pattern
			List<int> rowsT = new List<int>();
			List<int> rowsF = new List<int>();
			foreach (int r in partitionRows)
			{
				BitArray row = data.getRow(r);
				bool contains = true;
				foreach(int i in pattern.items)
				{
					if (!row.get(i))
					{
						contains = false;
						break;
					}
				}

				if (contains)
					rowsT.Add(r);
				else
					rowsF.Add(r);
			}
			Console.WriteLine(rowsT.Count + " " + rowsF.Count);
			//create two partitions
			Partition t = new Partition(data, trueRows, rowsT);
			Partition f = new Partition(data, trueRows, rowsF);
			List<Partition> splitresult = new List<Partition>();
			if (rowsT.Count == 0 || rowsF.Count == 0)
			{
				splitresult.Add(this);
				return splitresult;
			}

			//set pattern for splitnode
			splitNode.pattern = pattern;

			//create left subtree
			splitNode.left = f.splitNode;
			f.splitNode.parent = splitNode;

			//create right subtree
			splitNode.right = t.splitNode;
			t.splitNode.parent = splitNode;

			splitresult.Add(t);
			splitresult.Add(t);
			return splitresult;
		}

		public int Count
		{
			get
			{
				return partitionRows.Count;
			}
		}
	}
}
