﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Globalization;
using System.Threading;

namespace CommunityDetection
{
	public class FinalProgramV2 : ExpProgram
	{
		int numThreads;
		BeamBuilder builder;
		string descFile,generatorType,generatorArg;
		BinaryMatrix dm;
		string logdir;
		ThreadSafeQueue<Community> candidates;
		double minDist;
		StreamWriter jobwriter;
		bool doFirstEmRun;
		int remineIterations, beamWidth, maxPatternLength;
		bool countErrorsDouble;
		int maxEmDmIterations;
		Dictionary<CommunityStamp, NBClassifier> classifiers;
		Dictionary<CommunityStamp, CommunityStamp> classifiedTo;
		ThreadSafeQueue<CommunityStamp> emOnlyList;

		public FinalProgramV2()
		{
		}

		public override void loadArguments(string[] args)
		{
			descFile = args[0];
			numThreads = int.Parse(args[1]);
			generatorType = args[2];
			generatorArg = args[3];
			minDist = double.Parse(args[4], CultureInfo.InvariantCulture);
			doFirstEmRun = args[5].ToLower().Trim().Equals("true");
			string[] tmp = args[6].Split(' ');
			remineIterations = int.Parse(tmp[0]);
			beamWidth = int.Parse(tmp[1]);
			maxPatternLength = int.Parse(tmp[2]);
			countErrorsDouble = args[7].ToLower().Trim().Equals("true");
			maxEmDmIterations = int.Parse(args[8]);
		}

		public override void  start()
		{
			logdir = expdir + expName;
			jobwriter = new StreamWriter(logdir + ".csv");
			explog("*** Settings ***");
			explog("Experiment:");
			explog(" ;File"+":;" + DateTime.Now.ToString("dd-MM-yyyy HH:mm:ss") + ".csv");
			explog(" ;Program:;" + this.GetType().Name);
			explog(" ;Name:;" + expName);
			explog("Files:");
			explog(" ;Datadir:;" + datadir);
			explog(" ;Expdir:;" + expdir);
			explog(" ;Logdir:;" + expName + "/");
			explog("Graph:");
			explog(" ;Graphname;:" + graphName);
			explog(" ;|V|:;" + graph.Count);
			explog(" ;|E|:;" + graph.NumEdges);
			explog("Description Matrix:");

			if(descFile.ToLower() != "null")
			{
				startmsg("Reading description matrix");
				dm = new BinaryMatrixBitArrays(datadir+descFile);
				endmsg();
				explog(" ;File:; "+descFile);
				explog(" ;M:;" + dm.M);
				explog(" ;N:;" + dm.N);
				explog(" ;Density:;" + dm.Density);
			}
			else
			{
				explog(" ;File:; None");
			}

			//threads
			explog("Threads");
			explog(" ;Number:;" + numThreads);
			
			
			Directory.CreateDirectory(logdir);

			//-------------------candidates-----------------
			
			//generator
			CandidatesGenerator generator;
			
			if (generatorType.ToLower().Equals("sn"))
			{
				int sn = int.Parse(generatorArg);
				explog("Candidates generator:");
				explog(" ;Method:;Shared neighbours");
				explog(" ;MinDist:;" + sn);
				generator = new SharedNeighboursCG(sn);
			}
			else if (generatorType.ToLower().Equals("desc"))
			{
				string pattern = generatorArg;
				generator = new DescriptionCandidateGenerator(pattern); 
				explog("Candidates generator:");
				explog(" ;Method:;Description");
				explog(" ;Pattern given:;" + pattern);
				explog(" ;Pattern interpreted:"+((DescriptionCandidateGenerator) generator).Description);
				
			}
			else
			{
				string file = generatorArg;
				explog("Candidates generator:");
				explog(" ;Method:;Candidate file");
				explog(" ;File:;" + file);
				generator = new CandidateFileGenerator(datadir+file);
			}

			//generating candidates
			startmsg("Generating candidates");
			candidates = generator.generate(graph, dm);
			endmsg();
			Console.WriteLine("Generated " + candidates.Count + " candidates");
			explog("Candidates: ");
			explog(" ;Before filtering:;" + candidates.Count);

			//first EM run
			explog(" ;Do first EM run:;" + doFirstEmRun);
			if (doFirstEmRun)
			{
				startmsg("Running first EM run to filter double candidates");
				candidates = firstEmRun(candidates);
				endmsg();
				explog(" ;After first EM run:;" + candidates.Count);
			}

			Console.WriteLine("After filtering we have " + candidates.Count + " candidates left");

			//measure
			explog("Count between-errors double");
			explog("  ;Double:;" + countErrorsDouble);
			Community.measure = new CommunityGain(countErrorsDouble);
			explog("Max EMD iterations (0 is unlimited)");
			explog(" ;Iterations:;" + maxEmDmIterations);

			//classifier
			explog("Classifier");
			if (dm != null)
			{
				startmsg("Building beam classifier");
				builder = new BeamBuilder(dm, remineIterations, beamWidth, maxPatternLength);
				endmsg();
				explog(" ;Type:;Remine with beam search");
				explog(" ;Iterations:;" + remineIterations);
				explog(" ;Beam width:;" + beamWidth);
				explog(" ;Max Pattern length:;" + maxPatternLength);
			}
			else
			{
				explog(" ;Type:;None, no description matrix");
			}

			explog("");
			explog("*** Initial candidates ***");
			explog("ID;Size;Community gain;ICD;Nodes");
			startmsg("Writing candidates to file");
			int count = candidates.Count;
			while (count > 0)
			{
				Community c = candidates.Dequeue();
				explog(c.ID+";"+c.Size+";"+c.communityGain()+";"+c.intraClusterDensity()+";"+c.ToString());
				candidates.Enqueue(c);
				count--;
			}
			endmsg();
			explog("");


			explog("Result communities");
			explog("ID;iteration;Place;Size;Community gain;ICD;Modularity;Conductance;Inverse Conductance;#Patterns;#Attributes in classifier;Total pattern length;AVG Pattern length;Complexity;Next Action;Nodes");
			startmsg("Running EMDM steps");
			emdm();
			endmsg();

			if (builder != null)
			{
				startmsg("Writing all communities with Classifier");
				explog("");
				explog("");
				explog("Communities with classifier");
				explog("ID;Size;Community gain;ICD;Modularity;Conductance;Inverse Conductance;#Patterns;#Attributes in classifier;Total pattern length;AVG Pattern length;Complexity;Score;Nodes");

				List<KeyValuePair<Community, NBClassifier>> list = new List<KeyValuePair<Community, NBClassifier>>();
				foreach (KeyValuePair<CommunityStamp, NBClassifier> kv in classifiers)
					list.Add(new KeyValuePair<Community, NBClassifier>(new Community(graph, classifiedTo[kv.Key]), kv.Value));

				list = new List<KeyValuePair<Community, NBClassifier>>(list.OrderByDescending(x => x.Key.communityGain() * 1.0 / x.Value.Complexity));
				foreach (KeyValuePair<Community, NBClassifier> kv in list)
				{
					Community c = kv.Key;
					NBClassifier cfr = kv.Value;
					explog(c.ID + ";" + c.Size + ";" + c.communityGain() + ";" + c.intraClusterDensity() + ";" + c.modularity() + ";" + c.conductance() + ";" + c.inverseConductance() + ";" + cfr.getInformation() + ";" + (c.communityGain() * 1.0 / cfr.Complexity) + ";" + c.ToString());
				}
				endmsg();

				startmsg("Filtering communities");
				if (minDist > 0)
				{
					explog("");
					explog("");
					explog("Communities after filtering");
					explog("ID;Size;Community gain;ICD;Modularity;Conductance;Inverse Conductance;#Patterns;#Attributes in classifier;Total pattern length;AVG Pattern length;Complexity;Score;Nodes");
					List<KeyValuePair<Community, NBClassifier>> afterFiltering = new List<KeyValuePair<Community, NBClassifier>>();
					foreach (KeyValuePair<Community, NBClassifier> kv in list)
					{
						bool similar = false;
						foreach (KeyValuePair<Community, NBClassifier> kv2 in afterFiltering)
						{
							double dist = kv.Key.JaccardDistance(kv2.Key);
							if (dist < minDist)
							{
								similar = true;
								break;
							}
						}
						if (!similar)
						{
							Community c = kv.Key;
							NBClassifier cfr = kv.Value;
							explog(c.ID + ";" + c.Size + ";" + c.communityGain() + ";" + c.intraClusterDensity() + ";" + c.modularity() + ";" + c.conductance() + ";" + c.inverseConductance() + ";" + cfr.getInformation() + ";" + (c.communityGain() * 1.0 / cfr.Complexity) + ";" + c.ToString());
							afterFiltering.Add(kv);
						}
					}
				}
				endmsg();
			}
			else
			{
				Console.WriteLine("Filtering communities..");
				if (minDist > 0)
				{
					explog("");
					explog("");
					explog("Communities after filtering");
					explog("ID;Size;Community gain;ICD;Modularity;Conductance;Inverse Conductance;#Patterns;#Attributes in classifier;Total pattern length;AVG Pattern length;Complexity;Score;Nodes");

					startmsg("removing doubles from " + emOnlyList.Count + " communities");
					List<HashSet<int>> unique = new List<HashSet<int>>();
					HashSet<CommunityStamp> stampset = new HashSet<CommunityStamp>();
					foreach (CommunityStamp stamp in emOnlyList.UnsafeQueue)
						stampset.Add(stamp);
					List<HashSet<int>> setlist = new List<HashSet<int>>();
					foreach (CommunityStamp stamp in stampset)
						setlist.Add(new HashSet<int>(stamp.Vertices));
					endmsg();
					
					startmsg("filtering on mindist the remaining "+stampset.Count+" communities");
					foreach (HashSet<int> stamp in setlist)
					{

						bool similar = false;
						foreach (HashSet<int> stamp2 in unique)
						{
							int m01 = stamp2.Except(stamp).Count();
							int m10 = stamp.Except(stamp2).Count();
							int m11 = stamp.Intersect(stamp2).Count();
							int m00 = graph.Count - m01 - m10 - m11;
							double jd = ((double)(m01 + m10)) / ((double)(m01 + m10 + m11));
							if (jd < minDist)
							{
								similar = true;
								break;
							}
						}
						if (!similar)
						{
							unique.Add(stamp);
							Community c = new Community(graph);
							foreach (int v in stamp)
								c.AddVertice(v);
							explog(c.ID + ";" + c.Size + ";" + c.communityGain() + ";" + c.intraClusterDensity() + ";" + c.modularity() + ";" + c.conductance() + ";" + c.inverseConductance() + "; NVT ; NVT ;" + c.ToString());
						}
					}
					endmsg();
					Console.WriteLine("After filtering: " + unique.Count);
				}
			}
		}

		public void explog(string msg)
		{
			lock (jobwriter)
			{
				jobwriter.WriteLine(msg);
				jobwriter.Flush();
			}
		}

		public void emdm()
		{
			List<Thread> threads = new List<Thread>(numThreads);
			classifiers = new Dictionary<CommunityStamp, NBClassifier>();
			classifiedTo = new Dictionary<CommunityStamp, CommunityStamp>();
			threads = new List<Thread>();
			emOnlyList = new ThreadSafeQueue<CommunityStamp>();
			Console.WriteLine();
			for (int i = 0; i < numThreads; i++)
			{
				EmDmRunV2 run = new EmDmRunV2(this, builder, candidates, graph, dm, logdir, maxEmDmIterations, classifiers, classifiedTo, emOnlyList);
				Thread t = new Thread(run.run);
				threads.Add(t);
				t.Start();
			}
			foreach (Thread t in threads)
				t.Join();

		}

		public ThreadSafeQueue<Community> firstEmRun(ThreadSafeQueue<Community> input)
		{
			CommunityStampSet stampSet = new CommunityStampSet();
			List<Thread> threads = new List<Thread>(numThreads);
			ThreadSafeQueue<Community> output = new ThreadSafeQueue<Community>();
			Dictionary<int, HashSet<int>> original = new Dictionary<int, HashSet<int>>();
			
			//save all starting sets
			int count = input.Count;
			while (count > 0)
			{
				Community c = input.Dequeue();
				original[c.ID] = new HashSet<int>(c.getVertices().Keys);
				input.Enqueue(c);
				count--;
			}

			//run EM run on all sets
			threads = new List<Thread>();
			for (int i = 0; i < numThreads; i++)
			{
				EmRun emRun = new EmRun(stampSet, input, output);
				Thread t = new Thread(emRun.run);
				threads.Add(t);
				t.Start();
			}
			//wait for them to finish
			foreach (Thread t in threads)
				t.Join();

			//restore communities to original vertices;
			count = output.Count;
			while (count > 0)
			{
				Community c = output.Dequeue();
				HashSet<int> current = new HashSet<int>(c.getVertices().Keys);
				
				//clear
				foreach (int v in current)
					c.RemoveVertice(v);
				
				//add starting vertices
				foreach (int v in original[c.ID])
					c.AddVertice(v);

				output.Enqueue(c);
				count--;
			}
			return output;
		}

		public ThreadSafeQueue<Community> filterCandidates(ThreadSafeQueue<Community> input)
		{
			ThreadSafeQueue<Community> output = new ThreadSafeQueue<Community>();
			bool similar;
			while (!input.Empty())
			{
				similar = false;
				Community cin = input.Dequeue();
				foreach (Community cout in output.UnsafeQueue)
				{
					double dist = cin.JaccardDistance(cout);
					if (dist < minDist)
					{
						similar = true;
						break;
					}
				}
				if (!similar)
					output.Enqueue(cin);
			}
			return output;
		}
	}
}
