﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading;

namespace CommunityDetection
{
	class RemineParamsExperiment
	{
		ThreadSafeQueue<int[]> queue;
		List<Thread> threads;
		StreamWriter writer;
		BinaryMatrix data;
		HashSet<int> trueRows;
		string dir;

		public RemineParamsExperiment()
		{
			StreamReader config = new StreamReader("config.cfg");

			Console.WriteLine("Reading description");
			data = new BinaryMatrixBitArrays(config.ReadLine());
			Console.WriteLine("Reading graph");
			UndirectedGraph g = new UndirectedGraphList(config.ReadLine());
			Console.WriteLine("Graph with " + g.Count + " nodes created");
			Console.WriteLine("Creating community");
			Community c = new Community(g);
			c.AddVertice(101);
			while (c.improve()) ;
			trueRows = new HashSet<int>(c.getVertices().Keys);
			Console.WriteLine("Community with size " + c.Size + " created");

			queue = new ThreadSafeQueue<int[]>();
			dir = "results_" + DateTime.Now.ToString("yyyy_MM_dd_HHmmss");
			writer = new StreamWriter(dir+".csv");
			Directory.CreateDirectory(dir);
			log("Graph:");
			log("  V:;" + g.Count);
			log("  E:;" + g.NumEdges);
			log("");
			log("Tags:; " + data.N);
			log("");
			log("Community:");
			log("  size:;" + c.Size);
			log("  nodes:;"+ Tools.listToString(new List<int>(c.getVertices().Keys)));
			log("");
			log("Remine iterations;Beam;Maxlength;Duration;Precision;Recall;Patterns;Nodes internal;Nodes external;TP;TN;FP;FN");
			for (int r = 6; r > 0; r--)
			{
				int[] bs = new int []{ 2, 4, 8, 15};
				foreach (int b in bs.Reverse())
				{
					for (int l = 6; l > 0; l--)
					{
						int[] action = new int[3] { r, b, l };
						queue.Enqueue(action);
					}
				}
			}

			Console.WriteLine("Threads running, queue size: " + queue.Count);

			//creating threads
			threads = new List<Thread>();
			int numthreads = int.Parse(config.ReadLine());
			for (int i = 0; i < numthreads; i++)
			{
				Thread t = new Thread(this.run);
				threads.Add(t);
				t.Start();
			}
			config.Close();

			

			//wait for all threads to be finished
			foreach (Thread t in threads)
				t.Join();
			writer.Close();
		}


		public void run()
		{
			while(!queue.Empty())
			{
				try
				{
					
					int threadId = 0;
					foreach(Thread t in threads)
					{
						if(t == Thread.CurrentThread)
							break;
						threadId++;
					}

					int[] p = queue.Dequeue();
					string filename = dir + "/exp_" + p[0] + "_" + p[1] + "_" + p[2];
					Console.WriteLine("Thread " + threadId + " starts on r:" + p[0] +", b:"+ p[1]+", l:" + p[2]);
					TimeMeasure measure = new TimeMeasure();
					BeamBuilder builder = new BeamBuilder(data, p[0], p[1], p[2]);
					BeamClassifier classifier = (BeamClassifier) builder.build(trueRows);
					double duration = measure.getDuration();
					
					StreamWriter treeWriter = new StreamWriter(filename+"_tree.txt");
					classifier.root.print(0, treeWriter);
					treeWriter.Close();
					
					StreamWriter patternWriter = new StreamWriter(filename + "_patterns.csv");
					patternWriter.WriteLine("Pattern;Records containing pattern;Records containing pattern in_comm;value of pattern");
					List<KeyValuePair<BeamPattern,double[]>> patterns = classifier.getPatterns();
					foreach (KeyValuePair<BeamPattern, double[]> kv in patterns)
					{
						string line = "";
						foreach (int i in kv.Key.pattern)
							line += " " + i;
						line = line.Trim() + ";" + kv.Value[0] + ";" + kv.Value[1] + ";" + kv.Value[2];
						patternWriter.WriteLine(line);
					}
					patternWriter.Close();


					int[] q = classifier.quality(data, trueRows);
					double precision = classifier.precision(q);
					double recall = classifier.recall(q);
					int[] nodes = classifier.root.count();
					
					StreamWriter expWriter = new StreamWriter(filename + "_exp.txt");
					expWriter.WriteLine("Settings: ");
					expWriter.WriteLine("  Remine iterations: " + p[0]);
					expWriter.WriteLine("  Beam size: " + p[1]);
					expWriter.WriteLine("  Max pattern length: " + p[2]);
					expWriter.WriteLine();
					expWriter.WriteLine("Results:");
					expWriter.WriteLine("  Precision: " + precision);
					expWriter.WriteLine("  Recall: " + recall);
					expWriter.WriteLine("  TP/FP/TN/FN: "+q[0]+"/"+q[1]+"/"+q[2]+"/"+q[3]);
					expWriter.WriteLine("  Duration: " + duration);
					expWriter.WriteLine();
					expWriter.WriteLine("Classifier:");
					expWriter.WriteLine("  Internal nodes: " + nodes[0]);
					expWriter.WriteLine("  External nodes: " + nodes[1]);
					expWriter.WriteLine("  Patterns: " + patterns.Count);

					expWriter.Close();

					log(p[0]+";"+p[1]+";"+p[2]+";"+duration+";"+precision+";"+recall+";"+patterns.Count+";"+nodes[0]+";"+nodes[1]+";"+q[0]+";"+q[1]+";"+q[2]+";"+q[3]);
					Console.WriteLine("Thread " + threadId + " finished " + p[0] + ", b:" + p[1] + ", l:" + p[2] + " in " + duration + " seconds");
					Console.WriteLine("Queue size: " + queue.Count);
				}
				catch(Exception)
				{
					return;
				}
			}
		}

		public void log(string msg)
		{
			lock (writer)
			{
				writer.WriteLine(msg);
				writer.Flush();
			}
		}
	}
}
